"""Socialite Settings"""
from masonite.environment import env

DRIVERS = {
    "github": {
        "client_id": env("GITHUB_CLIENT_ID"),
        "client_secret": env("GITHUB_CLIENT_SECRET"),
        "redirect": "/auth/callback/github",
    },
    "gitlab": {
        "client_id": env("GITLAB_CLIENT_ID"),
        "client_secret": env("GITLAB_CLIENT_SECRET"),
        "redirect": "/auth/callback/gitlab",
    },
    "bitbucket": {
        "client_id": env("BITBUCKET_CLIENT_ID"),
        "client_secret": env("BITBUCKET_CLIENT_SECRET"),
        "redirect": "/auth/callback/bitbucket",
    },
    "google": {
        "client_id": env("GOOGLE_CLIENT_ID"),
        "client_secret": env("GOOGLE_CLIENT_SECRET"),
        "redirect": "/auth/callback/google",
    },
    "apple": {
        "client_id": env("APPLE_CLIENT_ID"),
        "client_secret": env("APPLE_CLIENT_SECRET"),
        "redirect": "/auth/callback/apple",
    },
    "facebook": {
        "client_id": env("FACEBOOK_CLIENT_ID"),
        "client_secret": env("FACEBOOK_CLIENT_SECRET"),
        "redirect": "/auth/callback/facebook",
    },
}



"""Cache Config"""
import os

DISKS = {
    "default": "local",
    "local": {
        "driver": "file",
        "path": os.path.join(os.getcwd(), "tests/integrations/storage/uploads/"),
    },
    "s3": {
        "driver": "s3",
        "client": os.getenv("AWS_CLIENT"),
        "secret": os.getenv("AWS_SECRET"),
        "bucket": os.getenv("AWS_BUCKET"),
    },
}



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .drivers.BaseDriver import BaseDriver

class OAuth:
    def add_driver(name: str, driver: "BaseDriver"):
        """Register a new driver for OAuth2."""
        ...
    def set_configuration(config: dict) -> OAuth:
        """Set configuration for OAuth2."""
        ...
    def driver(name: str) -> "BaseDriver":
        """Get OAuth2 instance for given driver."""
        ...
    def get_config_options(driver: str = None) -> dict:
        """Get configuration options for a given driver or the default driver."""
        ...



from .BaseDriver import BaseDriver


class GitlabDriver(BaseDriver):
    def get_default_scopes(self):
        return ["read_user"]

    def get_auth_url(self):
        return "https://gitlab.com/oauth/authorize"

    def get_token_url(self):
        return "https://gitlab.com/oauth/token"

    def get_email_url(self):
        return ""

    def get_user_url(self):
        return "https://gitlab.com/api/v4/user"

    def get_request_options(self, token):
        return {"headers": {"Authorization": f"Bearer {token}"}}

    def map_user_data(self, data):
        return {
            "id": data["id"],
            "nickname": data["username"],
            "name": data["name"],
            "email": data["email"],
            "avatar": data["avatar_url"],
        }

    def revoke(self, token):
        # https://docs.gitlab.com/ee/api/oauth2.html#revoke-a-token
        response = self.perform_basic_request(
            "post",
            "https://gitlab.com/oauth/revoke",
            json={"token": token},
            headers={"Accept": "application/json"},
        )
        if response.status_code == 200:
            return True
        else:
            return False



"""A SocialiteProvider Service Provider."""
from masonite.configuration import config
from masonite.packages import PackageProvider

from ..OAuth import OAuth
from ..drivers import (
    GithubDriver,
    GitlabDriver,
    BitbucketDriver,
    GoogleDriver,
    AppleDriver,
    FacebookDriver,
)


class OAuthProvider(PackageProvider):
    """ServiceProvider for OAuth package"""

    def configure(self):
        (self.root("masonite/oauth").name("oauth").config("config/oauth.py", publish=True))

        oauth = OAuth(self.application).set_configuration(config("oauth.drivers"))
        oauth.add_driver("github", GithubDriver(self.application))
        oauth.add_driver("gitlab", GitlabDriver(self.application))
        oauth.add_driver("bitbucket", BitbucketDriver(self.application))
        oauth.add_driver("google", GoogleDriver(self.application))
        oauth.add_driver("apple", AppleDriver(self.application))
        oauth.add_driver("facebook", FacebookDriver(self.application))
        self.application.bind("oauth", oauth)

    def boot(self):
        """Boots services required by the container."""
        pass



""" Database Settings """
from masonite.environment import LoadEnvironment, env
from masoniteorm.connections import ConnectionResolver

"""
|--------------------------------------------------------------------------
| Load Environment Variables
|--------------------------------------------------------------------------
|
| Loads in the environment variables when this page is imported.
|
"""

LoadEnvironment()

"""
The connections here don't determine the database but determine the "connection".
They can be named whatever you want.
"""

DATABASES = {
    "default": env("DB_CONNECTION", "sqlite"),
    "sqlite": {
        "driver": "sqlite",
        "database": env("SQLITE_DB_DATABASE", "masonite.sqlite3"),
        "prefix": "",
    },
    "mysql": {
        "driver": "mysql",
        "host": env("DB_HOST"),
        "user": env("DB_USERNAME"),
        "password": env("DB_PASSWORD"),
        "database": env("DB_DATABASE"),
        "port": env("DB_PORT"),
        "prefix": "",
        "grammar": "mysql",
        "options": {
            "charset": "utf8mb4",
        },
    },
    "postgres": {
        "driver": "postgres",
        "host": env("DB_HOST"),
        "user": env("DB_USERNAME"),
        "password": env("DB_PASSWORD"),
        "database": env("DB_DATABASE"),
        "port": env("DB_PORT"),
        "prefix": "",
        "grammar": "postgres",
    },
    "mssql": {
        "driver": "mssql",
        "host": env("MSSQL_DATABASE_HOST"),
        "user": env("MSSQL_DATABASE_USER"),
        "password": env("MSSQL_DATABASE_PASSWORD"),
        "database": env("MSSQL_DATABASE_DATABASE"),
        "port": env("MSSQL_DATABASE_PORT"),
        "prefix": "",
    },
}

DB = ConnectionResolver().set_connection_details(DATABASES)



from .BaseDriver import BaseDriver


class AppleDriver(BaseDriver):
    def get_default_scopes(self):
        return ["name", "email"]

    def get_auth_url(self):
        return "https://appleid.apple.com/auth/authorize"

    def get_token_url(self):
        return "https://appleid.apple.com/auth/token"

    def map_user_data(self, data):
        return {
            "id": data["sub"],
            "nickname": data["nickname"],
            "name": data["name"],
            "email": data["email"],
            "avatar": data["picture"],
        }



import os


DRIVERS = {
    "default": "smtp",
    "smtp": {
        "host": os.getenv("MAIL_HOST"),
        "port": os.getenv("MAIL_PORT"),
        "username": os.getenv("MAIL_USERNAME"),
        "password": os.getenv("MAIL_PASSWORD"),
    },
    "mailgun": {
        "domain": os.getenv("MAILGUN_DOMAIN"),
        "secret": os.getenv("MAILGUN_SECRET"),
    },
}



from masonite.providers import (
    RouteProvider,
    FrameworkProvider,
    ViewProvider,
    WhitenoiseProvider,
    ExceptionProvider,
    MailProvider,
    SessionProvider,
    QueueProvider,
    CacheProvider,
    EventProvider,
    StorageProvider,
    HelpersProvider,
    BroadcastProvider,
    AuthenticationProvider,
    AuthorizationProvider,
)
from masoniteorm.providers import ORMProvider

# register local package
from src.masonite.oauth import OAuthProvider

PROVIDERS = [
    FrameworkProvider,
    HelpersProvider,
    RouteProvider,
    ViewProvider,
    WhitenoiseProvider,
    ExceptionProvider,
    MailProvider,
    SessionProvider,
    CacheProvider,
    QueueProvider,
    EventProvider,
    StorageProvider,
    BroadcastProvider,
    AuthenticationProvider,
    AuthorizationProvider,
    ORMProvider,
    OAuthProvider,
]



""" Web Routes """
from masonite.routes import Route

ROUTES = [
    Route.get("/", "WelcomeController@show").name("welcome"),
    # Step 1
    Route.get("/auth/redirect/@provider", "WelcomeController@auth").name("auth.redirect"),
    # Step 2
    Route.get("/auth/callback/@provider", "WelcomeController@callback").name("auth.callback"),
]



class OAuthUser:
    def __init__(self):
        self.id = None
        self.name = None
        self.nickname = None
        self.email = None
        self.avatar = None
        self.token = None
        self.refresh_token = None
        self.expires_in = None

    def set_token(self, token):
        self.token = token
        return self

    def set_refresh_token(self, token):
        self.refresh_token = token
        return self

    def set_expires_in(self, expires_in):
        self.expires_in = expires_in
        return self

    def build(self, raw_data):
        self.id = raw_data.get("id", None)
        self.name = raw_data.get("name", None)
        self.nickname = raw_data.get("nickname", None)
        self.email = raw_data.get("email", None)
        self.avatar = raw_data.get("avatar", None)
        self.refresh_token = raw_data.get("refresh_token", None)
        self.expires_in = raw_data.get("expires_in", None)
        return self



import requests
from urllib.parse import urlparse, parse_qs
from masonite.tests import TestCase
from masonite.routes import Route


class TestProviders(TestCase):
    def setUp(self):
        super().setUp()
        self.setRoutes(
            # Step 1
            Route.get("/auth/redirect/@provider", "WelcomeController@auth").name("auth.redirect"),
            # Step 2
            Route.get("/auth/callback/@provider", "WelcomeController@callback").name(
                "auth.callback"
            ),
        )

    def test_apple(self):
        response = self.get("/auth/redirect/apple").assertRedirect()
        redirect_url = response.response.header("Location")
        data = urlparse(redirect_url)
        redirect_params = parse_qs(data.query)
        redirect_uri = redirect_params.get("redirect_uri")[0]
        state = redirect_params.get("state")[0]
        assert data.netloc == "appleid.apple.com"
        assert redirect_uri == "http://localhost:8000/auth/callback/apple"
        response.assertSessionHas("state", state)
        provider_response = requests.get(redirect_url)
        assert provider_response.status_code == 200

    def test_github(self):
        response = self.get("/auth/redirect/github").assertRedirect()
        redirect_url = response.response.header("Location")
        data = urlparse(redirect_url)
        redirect_params = parse_qs(data.query)
        redirect_uri = redirect_params.get("redirect_uri")[0]
        state = redirect_params.get("state")[0]
        assert data.netloc == "github.com"
        assert redirect_uri == "http://localhost:8000/auth/callback/github"
        response.assertSessionHas("state", state)
        provider_response = requests.get(redirect_url)
        assert provider_response.status_code == 200

    def test_gitlab(self):
        response = self.get("/auth/redirect/gitlab").assertRedirect()
        redirect_url = response.response.header("Location")
        print(redirect_url)
        data = urlparse(redirect_url)
        redirect_params = parse_qs(data.query)
        redirect_uri = redirect_params.get("redirect_uri")[0]
        state = redirect_params.get("state")[0]
        assert data.netloc == "gitlab.com"
        assert redirect_uri == "http://localhost:8000/auth/callback/gitlab"
        response.assertSessionHas("state", state)
        # browser checks prevent us to check url during test => 503

    def test_google(self):
        response = self.get("/auth/redirect/google").assertRedirect()
        redirect_url = response.response.header("Location")
        data = urlparse(redirect_url)
        redirect_params = parse_qs(data.query)
        redirect_uri = redirect_params.get("redirect_uri")[0]
        state = redirect_params.get("state")[0]
        assert data.netloc == "accounts.google.com"
        assert redirect_uri == "http://localhost:8000/auth/callback/google"
        response.assertSessionHas("state", state)
        provider_response = requests.get(redirect_url)
        assert provider_response.status_code == 200

    def test_bitbucket(self):
        response = self.get("/auth/redirect/bitbucket").assertRedirect()
        redirect_url = response.response.header("Location")
        data = urlparse(redirect_url)
        redirect_params = parse_qs(data.query)
        redirect_uri = redirect_params.get("redirect_uri")[0]
        state = redirect_params.get("state")[0]
        assert data.netloc == "bitbucket.org"
        assert redirect_uri == "http://localhost:8000/auth/callback/bitbucket"
        response.assertSessionHas("state", state)

    # gives an error on github CI...
    # def test_facebook(self):
    #     response = self.get("/auth/redirect/facebook").assertRedirect()
    #     redirect_url = response.response.header("Location")
    #     data = urlparse(redirect_url)
    #     redirect_params = parse_qs(data.query)
    #     redirect_uri = redirect_params.get("redirect_uri")[0]
    #     state = redirect_params.get("state")[0]
    #     assert data.netloc == "www.facebook.com"
    #     assert redirect_uri == "http://localhost:8000/auth/callback/facebook"
    #     response.assertSessionHas("state", state)
    #     provider_response = requests.get(redirect_url)
    #     assert provider_response.status_code == 200



import requests

from .BaseDriver import BaseDriver


class GoogleDriver(BaseDriver):
    def get_default_scopes(self):
        return ["openid", "profile", "email"]

    def get_auth_url(self):
        return "https://accounts.google.com/o/oauth2/auth"

    def get_token_url(self):
        return "https://www.googleapis.com/oauth2/v4/token"

    def get_user_url(self):
        return "https://www.googleapis.com/oauth2/v3/userinfo"

    def get_request_options(self, token):
        return {
            "headers": {"Authorization": f"Bearer {token}", "Accept": "application/json"},
            "query": {"prettyPrint": "false"},
        }

    def map_user_data(self, data):
        return {
            "id": data["sub"],
            "nickname": data["nickname"],
            "name": data["name"],
            "email": data["email"],
            "avatar": data["picture"],
        }

    def revoke(self, token):
        # https://developers.google.com/identity/protocols/oauth2/javascript-implicit-flow#oauth-2.0-endpoints_6
        response = requests.post(
            f"https://oauth2.googleapis.com/revoke?token={token}",
            headers={"Content-Type": "application/x-www-form-urlencoded"},
        )
        if response.status_code == 200:
            return True
        else:
            return False



class OAuth:
    def __init__(self, application, driver_config=None):
        self.application = application
        self.drivers = {}
        self.driver_config = driver_config or {}

    def add_driver(self, name, driver):
        self.drivers.update({name: driver})

    def set_configuration(self, config):
        self.driver_config = config
        return self

    def driver(self, name):
        try:
            selected_driver = self.drivers[name]
        except KeyError:
            raise Exception(f"The driver {name} has not been registered as an OAuth driver.")
        return selected_driver.set_options(self.get_config_options(name))

    def get_config_options(self, driver=None):
        if driver is None:
            return self.driver_config.get(self.driver_config.get("default"), {})
        return self.driver_config.get(driver, {})



import os
from masonite.tests import TestCase


class TestInstallPackage(TestCase):
    def test_publish_package(self):
        (
            self.craft("package:publish", "oauth")
            .assertSuccess()
            .assertOutputContains("Config")
            .assertOutputContains("tests/integrations/config/oauth.py")
        )
        assert os.path.isfile("tests/integrations/config/oauth.py")
        os.remove("tests/integrations/config/oauth.py")



from .BaseDriver import BaseDriver


class BitbucketDriver(BaseDriver):
    """
    doc
    https://confluence.atlassian.com/bitbucketserver/bitbucket-oauth-2-0-provider-api-1108483661.html
    https://developer.atlassian.com/cloud/bitbucket/oauth-2/
    """

    def get_default_scopes(self):
        # https://developer.atlassian.com/cloud/bitbucket/bitbucket-cloud-rest-api-scopes/
        return ["email"]

    def get_auth_url(self):
        return "https://bitbucket.org/site/oauth2/authorize"

    def get_token_url(self):
        return "https://bitbucket.org/site/oauth2/access_token"

    def get_user_url(self):
        return "https://api.bitbucket.org/2.0/user"

    def get_email_url(self):
        return "https://api.bitbucket.org/2.0/user/emails"

    def get_request_options(self, token):
        return {"headers": {"Authorization": f"Bearer {token}"}}

    def get_email_by_token(self, token):
        email = ""
        if self.has_scope("email"):
            emails_data = super().get_email_by_token(token)
            for email_data in emails_data.get("values", []):
                if (
                    email_data["is_primary"]
                    and email_data["is_confirmed"]
                    and email_data["type"] == "email"
                ):
                    email = email_data["email"]
        return email

    def map_user_data(self, data):
        return {
            "id": data["uuid"],
            "nickname": data["username"],
            "name": data["display_name"],
            "email": data.get("email", ""),
            "avatar": data["links"]["avatar"]["href"],
        }

    def revoke(self, token):
        response = self.perform_basic_request(
            "post",
            "https://bitbucket.org/site/oauth2/revoke",
            json={"token": token},
            headers={"Accept": "application/json"},
        )
        if response.status_code == 200:
            return True
        else:
            return False



import requests
from urllib.parse import urlencode

from masonite.exceptions import RouteNotFoundException
from masonite.facades import Url
from masonite.utils.str import random_string

from ..OAuthUser import OAuthUser


class BaseDriver:
    def __init__(self, application):
        self.application = application
        self.options = {}

        self._is_stateless = False
        self._scopes = self.get_default_scopes()
        self._data = {}
        self._scope_separator = ","

    def set_options(self, options):
        self.options = options
        return self

    def get_absolute_redirect_uri(self):
        redirect_route_or_url = self.options.get("redirect")
        if not redirect_route_or_url:
            raise Exception("You should specify a redirect route name or URI")
        # convert to route with name
        try:
            redirect_url = self.application.make("router").route(redirect_route_or_url)
        except RouteNotFoundException:
            redirect_url = redirect_route_or_url
        return Url.url(redirect_url)

    def get_auth_params(self, state=None):
        params = {
            "client_id": self.options.get("client_id"),
            "redirect_uri": self.get_absolute_redirect_uri(),
            "scope": self.format_scopes(),
            "response_type": "code",
        }
        if self.use_state():
            params["state"] = state
        # add custom params
        params.update(self._data)

        return params

    def build_auth_url(self, state):
        params = self.get_auth_params(state)
        base_url = self.get_auth_url()
        query = urlencode(params)
        return f"{base_url}?{query}"

    def redirect(self, state=None):
        if self.use_state():
            # let user provide their own generated state
            state = state or self.get_state()
            # self.application.make("session").set("state", state)
            self.application.make("request").session.set("state", state)
        else:
            state = None
        authorization_url = self.build_auth_url(state)

        return self.application.make("response").redirect(location=authorization_url)

    def get_token_fields(self, code):
        fields = {
            "grant_type": "authorization_code",
            "client_id": self.options.get("client_id"),
            "client_secret": self.options.get("client_secret"),
            "code": code,
            "redirect_uri": self.get_absolute_redirect_uri(),
        }
        return fields

    def get_token(self):
        code = self.application.make("request").input("code")
        data = self.get_token_fields(code)
        response = requests.post(
            self.get_token_url(), data, headers={"Accept": "application/json"}
        )
        data = response.json()
        if response.status_code != 200:
            raise Exception("error")
        return data

    def stateless(self):
        self._is_stateless = True
        return self

    def use_state(self):
        return not self._is_stateless

    def get_state(self):
        return random_string(40)

    def scopes(self, scopes_list):
        self._scopes.extend(scopes_list)
        return self

    def set_scopes(self, scopes_list):
        self._scopes = scopes_list
        return self

    def has_scope(self, scope):
        return scope in self._scopes

    def format_scopes(self):
        self._scopes = list(set(self._scopes))
        # order list of scopes alphabetically
        self._scopes.sort()
        #
        return self._scope_separator.join(self._scopes)

    def with_data(self, data):
        """Add optional parameters in the redirect request that the provider might support.
        data should be a dict that will be converted into URL GET params."""
        self._data = data
        return self

    def has_invalid_state(self):
        if self._is_stateless:
            return False
        state = self.application.make("request").session.get("state")
        return state != self.application.make("request").input("state")

    def get_default_scopes(self) -> list:
        return []

    def reset_scopes(self):
        self._scopes = self.get_default_scopes()

    def get_auth_url(self):
        raise NotImplementedError()

    def get_token_url(self):
        raise NotImplementedError()

    def get_refresh_token_url(self):
        return self.get_token_url()

    def get_revoke_token_url(self):
        raise NotImplementedError()

    def get_user_url(self):
        raise NotImplementedError()

    def get_email_url(self):
        raise NotImplementedError()

    def get_request_options(self, *args):
        raise NotImplementedError()

    def map_user_data(self, data) -> "OAuthUser":
        raise NotImplementedError()

    def revoke(self, token) -> bool:
        raise NotImplementedError()

    def get_email_by_token(self, token):
        """E-mail is often not send directly with user info, a subsequent request needs to be
        made in order to fetch user e-mail (if scopes allow it)."""
        if self.get_email_url():
            response = requests.get(self.get_email_url(), **self.get_request_options(token))
            return response.json()
        else:
            return {}

    def user(self) -> "OAuthUser":
        if self.has_invalid_state():
            raise Exception("Invalid state")
        data = self.get_token()
        token = data.get("access_token")
        user = self.user_from_token(token)
        user.set_refresh_token(data.get("refresh_token")).set_expires_in(data.get("expires_in"))
        return user

    def user_from_token(self, token: str) -> "OAuthUser":
        response = requests.get(self.get_user_url(), **self.get_request_options(token))
        data = response.json()
        if response.status_code != 200:
            raise Exception("Provider API error")
        email = self.get_email_by_token(token)
        if email:
            data.update({"email": email})
        user = OAuthUser().build(self.map_user_data(data))
        user.set_token(token)
        return user

    def refresh(self, refresh_token: str) -> "OAuthUser":
        params = {
            "client_id": self.options.get("client_id"),
            "client_secret": self.options.get("client_secret"),
            "refresh_token": refresh_token,
            "grant_type": "refresh_token",
            "redirect_uri": self.get_absolute_redirect_uri(),
        }
        query = urlencode(params)
        url = f"{self.get_refresh_token_url()}?{query}"
        response = requests.post(url, headers={"Accept": "application/json"})
        if response.status_code != 200:
            raise Exception("Provider API error")
        data = response.json()
        token = data.get("access_token")
        user = self.user_from_token(token)
        user.set_refresh_token(data.get("refresh_token")).set_expires_in(data.get("expires_in"))
        return user

    def perform_request(self, token, method, url, **options):
        """Perform an authenticated request to the API on behalf on the user to which the given
        token belongs with the token as a 'Bearer' authentication token."""
        request_options = {
            **self.get_request_options(token),
            **options,
        }
        response = requests.request(method, url, **request_options)
        return response

    def perform_basic_request(self, method, url, **options):
        """Perform an authenticated request to the API on behalf on the user to which the given
        token belongs with Basic HTTP authentication."""
        request_options = {
            "auth": requests.auth.HTTPBasicAuth(
                self.options.get("client_id"), self.options.get("client_secret")
            ),
            **options,
        }
        response = requests.request(method, url, **request_options)
        return response



# flake8: noqa: F401
from .GithubDriver import GithubDriver
from .GitlabDriver import GitlabDriver
from .BitbucketDriver import BitbucketDriver
from .GoogleDriver import GoogleDriver
from .AppleDriver import AppleDriver
from .FacebookDriver import FacebookDriver



from .BaseDriver import BaseDriver
from ..OAuthUser import OAuthUser


class GithubDriver(BaseDriver):
    def get_default_scopes(self):
        return ["user:email"]

    def get_auth_url(self):
        return "https://github.com/login/oauth/authorize"

    def get_token_url(self):
        return "https://github.com/login/oauth/access_token"

    def get_user_url(self):
        return "https://api.github.com/user"

    def get_email_url(self):
        return "https://api.github.com/user/emails"

    def get_request_options(self, token):
        # https://docs.github.com/en/rest/overview/resources-in-the-rest-api#current-version
        return {
            "headers": {
                "Authorization": f"token {token}",
                "Accept": "application/vnd.github.v3+json",
            }
        }

    def get_email_by_token(self, token):
        email = ""
        if self.has_scope("user:email"):
            emails_data = super().get_email_by_token(token)
            for email_data in emails_data:
                if email_data["primary"] and email_data["verified"]:
                    email = email_data["email"]
                    break
        return email

    def map_user_data(self, data):
        return {
            "id": data["id"],
            "nickname": data["login"],
            "name": data["name"],
            "email": data["email"],
            "avatar": data["avatar_url"],
        }

    def revoke(self, token):
        # https://docs.github.com/en/rest/reference/apps#delete-an-app-token"
        response = self.perform_basic_request(
            "delete",
            f"https://api.github.com/applications/{self.options.get('client_id')}/grant",
            json={"access_token": token},
            headers={"Accept": "application/vnd.github.v3+json"},
        )
        if response.status_code == 204:
            return True
        else:
            return False

    def user(self) -> "OAuthUser":
        """
        GitHub Oauth2 mechanism does not provide refresh token info from OAuth flow, a subsequent
        request needs to be done.
        https://docs.github.com/en/rest/reference/apps#check-a-token
        """

        if self.has_invalid_state():
            raise Exception("Invalid state")
        data = self.get_token()
        token = data.get("access_token")
        user = self.user_from_token(token)

        response = self.perform_basic_request(
            "post",
            f"https://api.github.com/applications/{self.options.get('client_id')}/token",
            json={"access_token": token},
            headers={"Accept": "application/vnd.github.v3+json"},
        )
        full_data = response.json()
        user.set_expires_in(full_data.get("expires_at"))
        return user



from urllib.parse import parse_qs, urlparse
from masonite.routes import Route
from masonite.tests import TestCase
from src.masonite.oauth import OAuth


class TestMethods(TestCase):
    def setUp(self):
        super().setUp()
        self.setRoutes(
            Route.get("/redirect-data", "WelcomeController@auth_with_data"),
            Route.get("/redirect-scopes", "WelcomeController@auth_with_scopes"),
        )

    def test_scopes(self):
        driver = OAuth.driver("github").scopes(["test:a"])
        assert driver._scopes == ["user:email", "test:a"]
        OAuth.driver("github").reset_scopes()

    def test_set_scopes(self):
        driver = OAuth.driver("github").set_scopes(["test:a"])
        assert driver._scopes == ["test:a"]
        OAuth.driver("github").reset_scopes()

    def test_with_data(self):
        driver = OAuth.driver("github").with_data({"user": "test"})
        assert driver._data == {"user": "test"}

    def test_request_with_additional_data(self):
        response = self.get("/redirect-data")
        redirect_url = response.response.header("Location")
        data = urlparse(redirect_url)
        redirect_params = parse_qs(data.query)
        assert redirect_params.get("framework")[0] == "masonite"

    def test_request_with_additional_scope(self):
        response = self.get("/redirect-scopes")
        redirect_url = response.response.header("Location")
        data = urlparse(redirect_url)
        redirect_params = parse_qs(data.query)

        scopes = redirect_params.get("scope")[0]
        assert "user:email" in scopes
        assert "admin:org" in scopes
        OAuth.driver("github").reset_scopes()



"""A WelcomeController Module."""
from masonite.views import View
from masonite.controllers import Controller
from masonite.request import Request

from src.masonite.oauth import OAuth


class WelcomeController(Controller):
    """WelcomeController Controller Class."""

    def show(self, view: View):
        return view.render("base")

    def auth(self, request: Request):
        provider = request.param("provider")
        return OAuth.driver(provider).redirect()

    def auth_with_data(self):
        return OAuth.driver("github").with_data({"framework": "masonite"}).redirect()

    def auth_with_scopes(self):
        return OAuth.driver("github").scopes(["admin:org"]).redirect()

    def callback(self, request: Request):
        provider = request.param("provider")
        user = OAuth.driver(provider).user()
        # user2 = OAuth.driver("github").user_from_token(user.token)
        # you now have a user object with data and a token
        return vars(user)



from .BaseDriver import BaseDriver


class FacebookDriver(BaseDriver):
    def get_default_scopes(self):
        return ["email"]

    def get_auth_url(self):
        return "https://www.facebook.com/dialog/oauth"

    def get_token_url(self):
        return "https://graph.facebook.com/oauth/access_token"

    def get_user_url(self):
        return "https://graph.facebook.com/me?"

    def get_request_options(self, token):
        return {
            "headers": {"Authorization": f"Bearer {token}", "Accept": "application/json"},
            "query": {"prettyPrint": "false"},
        }

    def map_user_data(self, data):
        return {
            "id": data["sub"],
            "nickname": data["nickname"],
            "name": data["name"],
            "email": data["email"],
            "avatar": data["picture"],
        }

    def revoke(self, token) -> bool:
        return
